def test_plot_predictions():
    X, y = make_line_data()
    regressor = LinearRegression()
    regressor.fit(X, y)

    fig, ax = plot_predictions(X, y, regressor, title="Test Plot")

    # Check that the function returns a Figure and Axes object
    assert isinstance(fig, plt.Figure)
    assert isinstance(ax, plt.Axes)

    # Check that the title of the plot is correct
    assert ax.get_title() == "Test Plot"

def test_plot_predictions_conformal():
    X, y = make_line_data()
    regressor = LinearRegression()
    regressor.fit(X, y)
    conformal = ConformalPredictor(regressor)
    conformal.fit(X, y)

    fig, ax = plot_predictions(X, y, conformal, conformal=True, title="Test Plot Conformal")

    # Check that the function returns a Figure and Axes object
    assert isinstance(fig, plt.Figure)
    assert isinstance(ax, plt.Axes)

    # Check that the title of the plot is correct
    assert ax.get_title() == "Test Plot Conformal"

def test_rmse():
    y_true = np.array([1, 2, 3, 4, 5])
    y_pred = np.array([1, 2, 3, 4, 5])
    assert rmse(y_true, y_pred) == 0

    y_pred = np.array([2, 3, 4, 5, 6])
    assert rmse(y_true, y_pred) == 1

def test_mae():
    y_true = np.array([1, 2, 3, 4, 5])
    y_pred = np.array([1, 2, 3, 4, 5])
    assert mae(y_true, y_pred) == 0

    y_pred = np.array([2, 3, 4, 5, 6])
    assert mae(y_true, y_pred) == 1

def test_coverage():
    y_true = np.array([1, 2, 3, 4, 5])
    y_pred_lower = np.array([0, 1, 2, 3, 4])
    y_pred_upper = np.array([2, 3, 4, 5, 6])
    assert coverage(y_true, y_pred_lower, y_pred_upper) == 1

    y_pred_upper = np.array([1, 2, 3, 3, 3])
    assert coverage(y_true, y_pred_lower, y_pred_upper) == 0.6

def test_sharpness():
    y_pred_lower = np.array([0, 1, 2, 3, 4])
    y_pred_upper = np.array([2, 3, 4, 5, 6])
    assert sharpness(y_pred_lower, y_pred_upper) == 2

    y_pred_upper = np.array([1, 2, 3, 4, 5])
    assert sharpness(y_pred_lower, y_pred_upper) == 1

def test_knn_regressor():
    # linear data
    for seed in [1, 2, 3, 4, 5]:
        X, y = make_line_data(random_seed=seed)
        # use default k
        knn = KNNRegressor()
        knn_sk = KNeighborsRegressor()
        knn.fit(X, y)
        knn_sk.fit(X, y)
        assert np.all(knn.predict(X) == knn_sk.predict(X))
        # use non-default k
        knn = KNNRegressor(k=10)
        knn_sk = KNeighborsRegressor(n_neighbors=10)
        knn.fit(X, y)
        knn_sk.fit(X, y)
        assert np.all(knn.predict(X) == knn_sk.predict(X))
    # sine data data
    for seed in [1, 2, 3, 4, 5]:
        X, y = make_sine_data(random_seed=seed)
        # use default k
        knn = KNNRegressor()
        knn_sk = KNeighborsRegressor()
        knn.fit(X, y)
        knn_sk.fit(X, y)
        assert np.all(knn.predict(X) == knn_sk.predict(X))
        # use non-default k
        knn = KNNRegressor(k=10)
        knn_sk = KNeighborsRegressor(n_neighbors=10)
        knn.fit(X, y)
        knn_sk.fit(X, y)
        assert np.all(knn.predict(X) == knn_sk.predict(X))

def test_linear_regressor():
    for seed in [1, 2, 3, 4, 5]:
        X, y = make_line_data(random_seed=seed)
        lr = LinearRegressor()
        lr_sk = LinearRegression()
        lr.fit(X, y)
        lr_sk.fit(X, y)
        assert np.sum(np.abs(lr.predict(X) - lr_sk.predict(X))) < 0.01

def test_conformal_predictor():
    # setup
    X_train, y_train = make_line_data(n_samples=200, random_seed=42)
    X_calibration, y_calibration = make_line_data(n_samples=1000, random_seed=42)
    X_test, y_test = make_line_data(n_samples=1000, random_seed=42)
    regressor = LinearRegression()
    regressor.fit(X_train, y_train)
    conformal_predictor = ConformalPredictor(regressor)
    conformal_predictor.fit(X_calibration, y_calibration)

    # test fit method
    assert conformal_predictor.scores is not None
    assert conformal_predictor.quantile is not None

    # test predict method
    y_pred, y_lower, y_upper = conformal_predictor.predict(X_test)
    assert np.all(y_pred == regressor.predict(X_test))
    assert np.all(y_lower <= y_pred)
    assert np.all(y_upper >= y_pred)

    # check coverage
    assert np.abs(coverage(y_test, y_lower, y_upper) - 0.95) < 0.025

    # check coverage with non-default alpha
    conformal_predictor = ConformalPredictor(regressor, alpha=0.2)
    conformal_predictor.fit(X_calibration, y_calibration)
    y_pred, y_lower, y_upper = conformal_predictor.predict(X_test)
    assert np.abs(coverage(y_test, y_lower, y_upper) - (1 - 0.20)) < 0.025

def test_make_line_data_default():
    X, y = make_line_data()
    assert len(X) == 100
    assert len(y) == 100
    assert X.shape == (100, 1)
    assert np.issubdtype(X.dtype, np.floating)
    assert np.issubdtype(y.dtype, np.floating)

def test_make_line_data_custom():
    X, y = make_line_data(n_samples=200, beta_0=2, beta_1=3, sd=4, X_low=-5, X_high=5, random_seed=1)
    assert len(X) == 200
    assert len(y) == 200
    assert X.shape == (200, 1)
    assert np.issubdtype(X.dtype, np.floating)
    assert np.issubdtype(y.dtype, np.floating)
    assert np.min(X) >= -5
    assert np.max(X) <= 5

def test_make_line_data_with_seed():
    X1, y1 = make_line_data(n_samples=200, beta_0=2, beta_1=3, sd=4, X_low=-5, X_high=5, random_seed=42)
    X2, y2 = make_line_data(n_samples=200, beta_0=2, beta_1=3, sd=4, X_low=-5, X_high=5, random_seed=42)
    X3, y3 = make_line_data(n_samples=200, beta_0=2, beta_1=3, sd=4, X_low=-5, X_high=5, random_seed=1)
    assert np.array_equal(X1, X2)
    assert np.array_equal(y1, y2)
    assert not np.array_equal(X1, X3)
    assert not np.array_equal(y1, y3)

def test_make_line_data_without_seed():
    X1, y1 = make_line_data(n_samples=200, beta_0=2, beta_1=3, sd=4, X_low=-5, X_high=5)
    X2, y2 = make_line_data(n_samples=200, beta_0=2, beta_1=3, sd=4, X_low=-5, X_high=5)
    assert not np.array_equal(X1, X2)
    assert not np.array_equal(y1, y2)

def test_make_sine_data_default():
    X, y = make_sine_data()
    assert len(X) == 100
    assert len(y) == 100
    assert X.shape == (100, 1)
    assert np.issubdtype(X.dtype, np.floating)
    assert np.issubdtype(y.dtype, np.floating)

def test_make_sine_data_custom():
    X, y = make_sine_data(n_samples=200, sd=4, X_low=-5, X_high=5, random_seed=1)
    assert len(X) == 200
    assert len(y) == 200
    assert X.shape == (200, 1)
    assert np.issubdtype(X.dtype, np.floating)
    assert np.issubdtype(y.dtype, np.floating)
    assert np.min(X) >= -5
    assert np.max(X) <= 5

def test_make_sine_data_with_seed():
    X1, y1 = make_sine_data(random_seed=42)
    X2, y2 = make_sine_data(random_seed=42)
    X3, y3 = make_sine_data(random_seed=1)
    assert np.array_equal(X1, X2)
    assert np.array_equal(y1, y2)
    assert not np.array_equal(X1, X3)
    assert not np.array_equal(y1, y3)

def test_make_sine_data_without_seed():
    X1, y1 = make_sine_data(n_samples=100)
    X2, y2 = make_sine_data(n_samples=100)
    assert not np.array_equal(X1, X2)
    assert not np.array_equal(y1, y2)

def test_split_data():
    X = np.array([[1, 2], [3, 4], [5, 6], [7, 8], [9, 10]])
    y = np.array([1, 2, 3, 4, 5])
    X_train, X_test, y_train, y_test = split_data(X, y, holdout_size=0.2, random_seed=42)

    assert X_train.shape[0] == 4
    assert X_test.shape[0] == 1
    assert y_train.shape[0] == 4
    assert y_test.shape[0] == 1

def test_split_data_with_seed():
    X = np.array([[1, 2], [3, 4], [5, 6], [7, 8], [9, 10]])
    y = np.array([1, 2, 3, 4, 5])
    X_train1, X_test1, y_train1, y_test1 = split_data(X, y, random_seed=42)
    X_train2, X_test2, y_train2, y_test2 = split_data(X, y, random_seed=42)
    X_train3, X_test3, y_train3, y_test3 = split_data(X, y, random_seed=1)

    assert np.array_equal(X_train1, X_train2)
    assert np.array_equal(X_test1, X_test2)
    assert np.array_equal(y_train1, y_train2)
    assert np.array_equal(y_test1, y_test2)
    assert not np.array_equal(y_train1, y_train3)
    assert not np.array_equal(X_train1, X_train3)

